﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;

namespace XamDataGridAddNewCancelSample
{
    /// <summary>
    /// Implementation of a dynamic data collection based on generic Collection&lt;T&gt;,<br />
    /// implementing INotifyCollectionChanged to notify listeners when items get added, removed or the whole list is refreshed.<br />
    /// In addition to <see cref="ObservableCollection{T}"/> this class implements the <see cref="IBindingList"/> interface.
    /// </summary>
    public class BindingObservableCollection<T> : ObservableCollection<T>, IBindingList
    {
        #region Constructors

        /// <summary>
        /// Initializes a new instance of BindingObservableCollection that is empty and has default initial capacity.
        /// </summary>
        /// <param name="newItemFactory">Optional factory to create a new item of type <see cref="T"/>. When omitted, the default constructor of <see cref="T"/> is used if available.</param>
        public BindingObservableCollection(Func<T> newItemFactory = null)
        {
            this.newItemFactory = newItemFactory
                                  ?? GetNewItemFactoryByReflection();
        }

        /// <summary>
        /// Initializes a new instance of the BindingObservableCollection class
        /// that contains element references copied from the specified list
        /// </summary>
        /// <param name="list">The list whose elements are copied to the new list.</param>
        /// <param name="newItemFactory">Optional factory to create a new item of type <see cref="T"/>. When omitted, the default constructor of <see cref="T"/> is used if available.</param>
        /// <remarks>
        /// The element-references are copied onto the BindingObservableCollection in the
        /// same order they are read by the enumerator of the list.
        /// </remarks>
        /// <exception cref="ArgumentNullException"> list is a null reference </exception>
        public BindingObservableCollection(List<T> list, Func<T> newItemFactory = null) : base(list)
        {
            this.newItemFactory = newItemFactory
                                  ?? GetNewItemFactoryByReflection();
        }

        /// <summary>
        /// Initializes a new instance of the BindingObservableCollection class
        /// that contains element references copied from the specified list
        /// </summary>
        /// <param name="collection">The collection whose elements are copied to the new list.</param>
        /// <param name="newItemFactory">Optional factory to create a new item of type <see cref="T"/>. When omitted, the default constructor of <see cref="T"/> is used if available.</param>
        /// <remarks>
        /// The element-references are copied onto the BindingObservableCollection in the
        /// same order they are read by the enumerator of the list.
        /// </remarks>
        /// <exception cref="ArgumentNullException"> list is a null reference </exception>
        public BindingObservableCollection(IEnumerable<T> collection, Func<T> newItemFactory = null) : base(collection)
        {
            this.newItemFactory = newItemFactory
                                  ?? GetNewItemFactoryByReflection();
        }

        /// <summary>
        /// Returns a new item factory method which calles the default constructor of <typeparam name="T"/>.<br />
        /// Returns null if the type does not have a default constructor.
        /// </summary>
        private Func<T> GetNewItemFactoryByReflection()
        {
            var itemType = typeof(T);
            var defaultConstructor = itemType.GetConstructor(Type.EmptyTypes);
            if (defaultConstructor == null)
                return null;

            return Activator.CreateInstance<T>;
        }

        /// <summary>
        /// This factory method creates new items for this collection.
        /// </summary>
        private readonly Func<T> newItemFactory = null;

        #endregion


        #region Properties

        // property backing fields
        private bool? allowNew = null;
        private bool allowEdit = true;
        private bool allowRemove = true;

        /// <summary>
        /// Ruft ab, ob der Liste mithilfe von <see cref="M:System.ComponentModel.IBindingList.AddNew"/> Elemente hinzugefügt werden können.
        /// </summary>
        /// <returns>
        /// true, wenn Sie der Liste mithilfe von <see cref="M:System.ComponentModel.IBindingList.AddNew"/> Elemente hinzufügen können, andernfalls false.
        /// </returns>
        public bool AllowNew
        {
            get
            {
                if (!allowNew.HasValue)
                    allowNew = (newItemFactory != null);

                return allowNew.Value;
            }
            set
            {
                if (allowNew.Equals(value))
                    return;

                if (value == true && (newItemFactory == null))
                {
                    var msg = string.Format("Cannot allow AddNew items to the collection containing items of type {0}.\n" + 
                                            "The type does not implement a default constructor.\n" + 
                                            "A NewItemFactory is missing also.", typeof(T).FullName);
                    throw new Exception(msg);
                }

                allowNew = value;
                OnListChanged(ListChangedType.Reset);
            }
        }

        /// <summary>
        /// Ruft ab, ob Elemente der Liste aktualisiert werden können.
        /// </summary>
        /// <returns>
        /// true, wenn Elemente der Liste aktualisiert werden können, andernfalls false.
        /// </returns>
        public bool AllowEdit
        {
            get { return allowEdit; }
            set
            {
                if (value == allowEdit)
                    return;

                allowEdit = value;
                OnListChanged(ListChangedType.Reset);
            }
        }

        /// <summary>
        /// Ruft ab, ob Elemente unter Verwendung von <see cref="M:System.Collections.IList.Remove(System.Object)"/> oder <see cref="M:System.Collections.IList.RemoveAt(System.Int32)"/> aus der Liste entfernt werden können.
        /// </summary>
        /// <returns>
        /// true, wenn Elemente aus der Liste entfernt werden können, andernfalls false.
        /// </returns>
        public bool AllowRemove
        {
            get { return allowRemove; }
            set
            {
                if (value == allowRemove)
                    return;

                allowRemove = value;
                OnListChanged(ListChangedType.Reset);
            }
        }

        /// <summary>
        /// Ruft ab, ob bei Änderungen der Liste oder eines Elements ein <see cref="E:System.ComponentModel.IBindingList.ListChanged"/>-Ereignis ausgelöst wird.<br />
        /// Diese Implementation unterstützt keine ChangeNotification.
        /// </summary>
        /// <returns>
        /// Gibt immer false zurück.
        /// </returns>
        public bool SupportsChangeNotification
        {
            get { return false; }
        }

        /// <summary>
        /// Ruft ab, ob die Liste die Suche mit der <see cref="M:System.ComponentModel.IBindingList.Find(System.ComponentModel.PropertyDescriptor,System.Object)"/>-Methode unterstützt.<br />
        /// Diese Implementation gibt grundsätzlich false zurück, da sie keine Suchoperationen unterstützt.
        /// </summary>
        /// <returns>
        /// true, wenn die Liste die Suche mit der <see cref="M:System.ComponentModel.IBindingList.Find(System.ComponentModel.PropertyDescriptor,System.Object)"/>-Methode unterstützt, andernfalls false.
        /// </returns>
        public bool SupportsSearching
        {
            get { return false; }
        }

        /// <summary>
        /// Ruft ab, ob die Liste die Sortierung unterstützt.<br />
        /// Diese Implementation gibt grundsätzlich false zurück, da sie keine Sortierung unterstützt.
        /// </summary>
        /// <returns>
        /// true, wenn die Liste die Sortierung unterstützt, andernfalls false.
        /// </returns>
        public bool SupportsSorting
        {
            get { return false; }
        }

        /// <summary>
        /// Ruft ab, ob die Element in der Liste sortiert werden.<br />
        /// Diese Implementation löst grundsätzlich eine <exception cref="T:System.NotSupportedException" /> aus, da sie keine Sortierung unterstützt.
        /// </summary>
        /// <returns>
        /// true, wenn <see cref="M:System.ComponentModel.IBindingList.ApplySort(System.ComponentModel.PropertyDescriptor,System.ComponentModel.ListSortDirection)"/> aufgerufen wurde und <see cref="M:System.ComponentModel.IBindingList.RemoveSort"/> nicht aufgerufen wurde, andernfalls false.
        /// </returns>
        /// <exception cref="T:System.NotSupportedException"><see cref="P:System.ComponentModel.IBindingList.SupportsSorting"/> ist false. </exception>
        public bool IsSorted
        {
            get { throw new NotSupportedException("This collection does not support sorting."); }
        }

        #endregion


        #region Public Methods

        /// <summary>
        /// Fügt der Liste ein neues Element hinzu.
        /// </summary>
        /// <returns>
        /// Das der Liste hinzugefügte Element.
        /// </returns>
        /// <exception cref="T:System.NotSupportedException"><see cref="P:System.ComponentModel.IBindingList.AllowNew"/> is false. </exception>
        public object AddNew()
        {
            if (!AllowNew)
                throw new Exception("Adding of new items is not supported on this collection.");

            // create a new item and add it to the list.
            var newItem = newItemFactory();
            newItem = InitializeNewItem(newItem);
            this.Add(newItem);
            return newItem;
        }

        #endregion


        #region Protected Methods

        /// <summary>
        /// Called by base class Collection&lt;T&gt; when the list is being cleared;
        /// raises a CollectionChanged event to any listeners.
        /// raises a OnListChanged(Reset) event.
        /// </summary>
        protected override void ClearItems()
        {
            base.ClearItems();
            OnListChanged(ListChangedType.Reset);
        }

        /// <summary>
        /// Called by base class Collection&lt;T&gt; when an item is removed from list;
        /// raises a CollectionChanged event to any listeners.
        /// raises a OnListChanged(ItemDeleted) event.
        /// </summary>
        protected override void RemoveItem(int index)
        {
            base.RemoveItem(index);
            OnListChanged(ListChangedType.ItemDeleted, index);
        }

        /// <summary>
        /// Called by base class Collection&lt;T&gt; when an item is added to list;
        /// raises a CollectionChanged event to any listeners.
        /// raises a OnListChanged(ItemAdded) event.
        /// </summary>
        protected override void InsertItem(int index, T item)
        {
            base.InsertItem(index, item);
            OnListChanged(ListChangedType.ItemAdded, index);
        }

        /// <summary>
        /// Called by base class Collection&lt;T&gt; when an item is set in list;
        /// raises a CollectionChanged event to any listeners.
        /// raises a OnListChanged(ItemChanged) event.
        /// </summary>
        protected override void SetItem(int index, T item)
        {
            base.SetItem(index, item);
            OnListChanged(ListChangedType.ItemChanged, index);
        }

        /// <summary>
        /// Called by base class ObservableCollection&lt;T&gt; when an item is to be moved within the list;
        /// raises a CollectionChanged event to any listeners.
        /// raises a OnListChanged(ItemMoved) event.
        /// </summary>
        protected override void MoveItem(int oldIndex, int newIndex)
        {
            base.MoveItem(oldIndex, newIndex);
            OnListChanged(ListChangedType.ItemMoved, oldIndex, newIndex);
        }

        /// <summary>
        /// Wird aufgerufen wenn die Collection selbst ein neues Element erzeugt hat.<br />
        /// Kann zur Initialisierung dieses Elementes genutzt werden.<br />
        /// Das Element darf auch ausgetauscht werden.
        /// </summary>
        /// <param name="item">Das von der Collection erzeugte neue Element</param>
        protected virtual T InitializeNewItem(T item)
        {
            return item;
        }

        #endregion


        #region Find/Search/Sort - not supported

        /// <summary>
        /// Ruft den <see cref="T:System.ComponentModel.PropertyDescriptor"/> ab, der für die Sortierung verwendet wird.<br />
        /// Wird von dieser Implementation nicht unterstützt!
        /// </summary>
        /// <returns>
        /// Der <see cref="T:System.ComponentModel.PropertyDescriptor"/>, der für die Sortierung verwendet wird.
        /// </returns>
        /// <exception cref="T:System.NotSupportedException"><see cref="P:System.ComponentModel.IBindingList.SupportsSorting"/> ist false. </exception>
        PropertyDescriptor IBindingList.SortProperty { get { return null; } }

        /// <summary>
        /// Ruft die Sortierrichtung ab.<br />
        /// Wird von dieser Implementation nicht unterstützt!
        /// </summary>
        /// <returns>
        /// Einer der <see cref="T:System.ComponentModel.ListSortDirection"/>-Werte.
        /// </returns>
        /// <exception cref="T:System.NotSupportedException"><see cref="P:System.ComponentModel.IBindingList.SupportsSorting"/> is false. </exception>
        ListSortDirection IBindingList.SortDirection { get{ return ListSortDirection.Ascending; } }

        /// <summary>
        /// Fügt den Indizes, die für die Suche verwendet werden, den <see cref="T:System.ComponentModel.PropertyDescriptor"/> hinzu.<br />
        /// Wird von dieser Implementation nicht unterstützt!
        /// </summary>
        /// <param name="property">Der <see cref="T:System.ComponentModel.PropertyDescriptor"/>, der den Indizes hinzugefügt werden soll, die für die Suche verwendet werden. </param>
        void IBindingList.AddIndex(PropertyDescriptor property)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Entfernt den <see cref="T:System.ComponentModel.PropertyDescriptor"/> aus den Indizes, die für die Suche verwendet werden.<br />
        /// Wird von dieser Implementation nicht unterstützt!
        /// </summary>
        /// <param name="property">Der <see cref="T:System.ComponentModel.PropertyDescriptor"/>, der aus den für die Suche verwendeten Indizes entfernt werden soll. </param>
        void IBindingList.RemoveIndex(PropertyDescriptor property)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Gibt den Index der Zeile mit dem angegebenen <see cref="T:System.ComponentModel.PropertyDescriptor"/> zurück.<br />
        /// Wird von dieser Implementation nicht unterstützt!
        /// </summary>
        /// <returns>
        /// Der Index der Zeile mit dem angegebenen <see cref="T:System.ComponentModel.PropertyDescriptor"/>.
        /// </returns>
        /// <param name="property">Der <see cref="T:System.ComponentModel.PropertyDescriptor"/>, in dem gesucht werden soll. </param><param name="key">Der Wert des <paramref name="property"/>-Parameters, nach dem gesucht werden soll. </param><exception cref="T:System.NotSupportedException"><see cref="P:System.ComponentModel.IBindingList.SupportsSearching"/> ist false. </exception>
        int IBindingList.Find(PropertyDescriptor property, object key)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Sortiert die Liste entsprechend einem <see cref="T:System.ComponentModel.PropertyDescriptor"/> und einer <see cref="T:System.ComponentModel.ListSortDirection"/>.<br />
        /// Wird von dieser Implementation nicht unterstützt!
        /// </summary>
        /// <param name="property">Der <see cref="T:System.ComponentModel.PropertyDescriptor"/>, nach dem sortiert werden soll. </param><param name="direction">One of the <see cref="T:System.ComponentModel.ListSortDirection"/> values. </param><exception cref="T:System.NotSupportedException"><see cref="P:System.ComponentModel.IBindingList.SupportsSorting"/> ist false. </exception>
        void IBindingList.ApplySort(PropertyDescriptor property, ListSortDirection direction)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Entfernt alle Sortierungen, die unter Verwendung von <see cref="M:System.ComponentModel.IBindingList.ApplySort(System.ComponentModel.PropertyDescriptor,System.ComponentModel.ListSortDirection)"/> angewendet wurden.<br />
        /// Wird von dieser Implementation nicht unterstützt!
        /// </summary>
        /// <exception cref="T:System.NotSupportedException"><see cref="P:System.ComponentModel.IBindingList.SupportsSorting"/> ist false. </exception>
        void IBindingList.RemoveSort()
        {
            throw new NotImplementedException();
        }

        #endregion


        #region Events

        /// <summary>
        /// Tritt ein, wenn die Liste oder ein Element der Liste geändert wird. <br />
        /// Wird von der <see cref="BindingObservableCollection{T}"/> Implementation nicht unterstützt.
        /// </summary>
        public event ListChangedEventHandler ListChanged;

        /// <summary>
        /// Raises the <see cref="ListChanged"/> event.
        /// </summary>
        /// <param name="e"></param>
        protected virtual void OnListChanged(ListChangedEventArgs e)
        {
            var handler = ListChanged;
            if (handler != null) handler(this, e);
        }

        /// <summary>
        /// Raises the <see cref="ListChanged"/> event.
        /// </summary>
        /// <param name="listChangedType">Kind of the change.</param>
        /// <param name="newIndex">Optional</param>
        /// <param name="oldIndex">Optional</param>
        protected virtual void OnListChanged(ListChangedType listChangedType, int? newIndex = null, int? oldIndex = null)
        {
            var handler = ListChanged;
            if (handler != null)
            {
                var args = new ListChangedEventArgs(listChangedType, newIndex ?? -1, oldIndex ?? -1);
                handler(this, args);
            }
        }

        #endregion
    }
}